/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class com_mesonbuild_JdkTest */

#ifndef _Included_com_mesonbuild_JdkTest
#define _Included_com_mesonbuild_JdkTest
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     com_mesonbuild_JdkTest
 * Method:    jdk_test
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_mesonbuild_JdkTest_jdk_1test
  (JNIEnv *, jclass);

#ifdef __cplusplus
}
#endif
#endif
