#include<alexandria.h>
#include<stdio.h>

int main(int argc, char **argv) {
    printf("You are standing outside the Great Library of Alexandria.\n");
    printf("You decide to go inside.\n\n");
    alexandria_visit();
    return 0;
}
